package FJsc.typechecker;

import java.io.IOException;
import java.io.Reader;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.io.InputStreamReader;
import java.util.HashSet;
import org.apache.bcel.generic.Type;
import org.apache.bcel.generic.ObjectType;

public class Parser{

    private StreamTokenizer st;
    private String className;

    public static String complete(String s){
	return s.equals("Object")?"java.lang.Object":s;
    }

    public Parser(StreamTokenizer st)throws IOException{
	this.st=st;
	st.next();
    }

    public Parser(String className)throws IOException,FileNotFoundException{
	this(new StreamTokenizer(new FileReader(className+".java")));
	this.className=className;
    }

    protected void check(int symType,String msg) throws ParserException,IOException {
	if (st.ttype!=symType)
	    throw new ParserException(msg);
	st.next();
    }


    protected static boolean beginExp(StreamTokenizer st){
	switch(st.ttype){ 
	case StreamTokenizer.TT_NEW:
	case StreamTokenizer.TT_WORD:
	case StreamTokenizer.TT_THIS:
	case StreamTokenizer.TT_NUMBER:
	case StreamTokenizer.TT_BEGIN_PAR:
	    return true;
	default:
	    return false;
	}
    }

    protected Exp parseExp()throws ParserException,IOException{
	Exp e;
	if(st.ttype==StreamTokenizer.TT_NUMBER)
	    return parseAtomExp();
	return parseFieldsMeths(parseAtomExp());
    }

    private Exp parseFieldsMeths(Exp e)throws ParserException,IOException{
	String n;
	while(st.ttype==StreamTokenizer.TT_DOT){
	    st.next();
	    n=parseName();
	    if(st.ttype==StreamTokenizer.TT_BEGIN_PAR){// method invocation
		st.next();
		e=new InvokeExp(e,n,parseArgs());
		check(StreamTokenizer.TT_END_PAR,"Expecting ')'");
	    }
	    else// field access
		e=new FieldExp(e,n);
	}
	return e;
    }

    protected Exp parseAtomExp()throws ParserException,IOException{
	Exp e=null;
	String n;
	switch(st.ttype){
	case StreamTokenizer.TT_NEW:
	    st.next();
	    n=parseName();
	    check(StreamTokenizer.TT_BEGIN_PAR,"Expecting '('");
	    e=new NewExp(n,parseArgs());
	    check(StreamTokenizer.TT_END_PAR,"Expecting ')'");
	    break;
	case StreamTokenizer.TT_WORD:
	case StreamTokenizer.TT_THIS:
	    e=new LocalVariableExp(st.sval);
	    st.next();
	    break;
	case StreamTokenizer.TT_NUMBER:
	    e=new NumLitExp(st.sval);
	    st.next();
	    break;
	case StreamTokenizer.TT_BEGIN_PAR:
	    ExtExp ee;
	    st.next();
	    if(st.ttype==StreamTokenizer.TT_INT){
		st.next();
		ee=IntExp.CONST;
	    }
            else
		ee=parseExp();
	    check(StreamTokenizer.TT_END_PAR,"Expecting ')'");
	    if(beginExp(st))
		e=new CastExp(ee.toType(),parseExp());
	    else		
		e=ee.toExp();
	    break;
	default:
	    throw new ParserException("Expecting either 'new', a name, or a number");  
	}
	return e;
    }

    protected AssignExp parseAssignExp()throws ParserException,IOException{
	check(StreamTokenizer.TT_THIS,"Expecting 'this'");
	check(StreamTokenizer.TT_DOT,"Expecting '.'");
	String field=parseName();
	check(StreamTokenizer.TT_ASSIGN,"Expecting '='");
	AssignExp ae=new AssignExp(new FieldExp(new LocalVariableExp("this"),field),parseExp());
	check(StreamTokenizer.TT_SEMICOLON,"Expecting ';'");
	return ae;
    }

    protected AssignExps parseAssignExps()throws ParserException,IOException{
	AssignExps aes=new AssignExps();
	while(st.ttype!=StreamTokenizer.TT_END_BLOCK)
	    aes.addLast(parseAssignExp());
	return aes;
    }

    protected Params parseParams()throws ParserException,IOException{
	Params ps=new Params();
	Param p;
	String n;
	HashSet hs=new HashSet();
	if(st.ttype==StreamTokenizer.TT_END_PAR)
	    return ps;
	while(true){
	    p=parseParam();
	    n=p.getName();
	    if(hs.contains(n))
		throw new ParserException("Parameter name "+n+" already used");
	    hs.add(n);
	    ps.addLast(p);
	    switch(st.ttype){
	    case StreamTokenizer.TT_END_PAR:
		return ps;
	    case StreamTokenizer.TT_COMMA:
		st.next();
		break;
	    default:
		throw new ParserException("Expecting either ',' or ')'");
	    }
	}
    }


    protected Param parseParam()throws ParserException,IOException{
	return new Param(parseType(),parseName());
    }

    protected Exps parseArgs()throws ParserException,IOException{
	Exps es=new Exps();
	if(st.ttype==StreamTokenizer.TT_END_PAR)
	    return es;
	while(true){
	    es.addLast(parseExp());
	    switch(st.ttype){
	    case StreamTokenizer.TT_END_PAR:
		return es;
	    case StreamTokenizer.TT_COMMA:
		st.next();
		break;
	    default:
		throw new ParserException("Expecting either ',' or ')'");
	    }
	}
    }

    protected Type parseType()throws ParserException,IOException{
	Type t;
	switch(st.ttype){
	case StreamTokenizer.TT_INT:
	    t=Type.INT;
	    st.next();
	    break;
	case StreamTokenizer.TT_WORD:
	    String s=parseName();
	    t=new ObjectType(complete(s));
	    break;
	default:
	    throw new ParserException("Expecting a type");
	}
	return t; 
    }

    protected String parseName()throws ParserException,IOException{
	if(st.ttype!=StreamTokenizer.TT_WORD)
	    throw new ParserException("Expecting a name");
	String s=st.sval;
	st.next();
	return s; 
    }

    protected String parseSimpleName()throws ParserException,IOException{
	if(st.ttype!=StreamTokenizer.TT_WORD)
	    throw new ParserException("Expecting a name");
	String s=st.sval;
	st.next();
	return s; 
    }

    protected void parseCons(SourceClassDec scd)throws ParserException,IOException{
	st.next();
	check(StreamTokenizer.TT_BEGIN_PAR,"Expecting '('");
	Params params=parseParams();
	check(StreamTokenizer.TT_END_PAR,"Expecting ')'");
	check(StreamTokenizer.TT_BEGIN_BLOCK,"Expecting '{'");
	AssignExps body=parseAssignExps();
	check(StreamTokenizer.TT_END_BLOCK,"Expecting '}'");
	Types ts=params.getTypes();
	if(scd.getCons().put(ts,new ConsDec(params,body))!=null)
	    throw new ParserException("Constructor "+scd.getName()+ts+" already defined");
    }

    protected void parseDec(SourceClassDec scd)throws ParserException,IOException{
	Type type=parseType();
	if(st.ttype==StreamTokenizer.TT_BEGIN_PAR && type.toString().equals(scd.getName())){// constructor
	    st.next();
	    Params params=parseParams();
	    check(StreamTokenizer.TT_END_PAR,"Expecting ')'");
	    check(StreamTokenizer.TT_BEGIN_BLOCK,"Expecting '{'");
	    AssignExps body=parseAssignExps();
	    check(StreamTokenizer.TT_END_BLOCK,"Expecting '}'");
	    Types ts=params.getTypes();
	    if(scd.getCons().put(ts,new ConsDec(params,body))!=null)
		throw new ParserException("Constructor "+scd.getName()+ts+" already defined");
	}
	else{// field or method 
	    String name=parseName();
	    if(st.ttype==StreamTokenizer.TT_SEMICOLON){// field
		st.next();
		if(scd.getFields().put(name,type)!=null)
		    throw new ParserException("Field "+name+" already defined in class "+scd.getName());
	    }
	    else if(st.ttype==StreamTokenizer.TT_BEGIN_PAR){// method
		st.next();
		Params params=parseParams();
		check(StreamTokenizer.TT_END_PAR,"Expecting ')'");
		check(StreamTokenizer.TT_BEGIN_BLOCK,"Expecting '{'");
		check(StreamTokenizer.TT_RETURN,"Expecting 'return'");
		Exp body=parseExp();
		check(StreamTokenizer.TT_SEMICOLON,"Expecting ';'");
		check(StreamTokenizer.TT_END_BLOCK,"Expecting '}'");
		MethSig ms=new MethSig(name,params.getTypes());
		if(scd.getMeths().put(ms,new MethDec(type,name,params,body))!=null)
		    throw new ParserException("Method "+name+" with signature "+ms+" already defined in class "+scd.getName());
	    }
	    else 
		throw new ParserException("Expecting either ';' or '('");
	}
    }

    protected void parseDecs(SourceClassDec scd)throws ParserException,IOException{
	while(st.ttype!=StreamTokenizer.TT_END_BLOCK)
	    parseDec(scd);
    }

    public SourceClassDec parseClassDec()throws ParserException,IOException{
	String name,parent;
	SourceClassDec classDec;
	ConsDecs cds=new ConsDecs();
	check(StreamTokenizer.TT_CLASS,"Expecting 'class'");
	name=parseSimpleName();
	//	if("java.lang.Object".equals(name))
	//  throw new ParserException("Class Object cannot be redefined");
	if(!className.equals(name))
	    throw new ParserException("Class "+name+" must be defined in file "+name+".java");
	check(StreamTokenizer.TT_EXTENDS,"Expecting 'extends'");
	parent=complete(parseName());
	check(StreamTokenizer.TT_BEGIN_BLOCK,"Expecting '{'");
	classDec=new SourceClassDec(name,parent,new FieldDecs(),cds,new MethDecs());
	parseDecs(classDec);
	check(StreamTokenizer.TT_END_BLOCK,"Expecting '}'");
	if(cds.isEmpty()) // must add default constructor
	    cds.put(new Types(),new ConsDec(new Params(),new AssignExps()));
	return classDec;
    }

} 

