package FJsc.typechecker;

import java.io.Reader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.util.HashMap;

public class StreamTokenizer extends java.io.StreamTokenizer{

    protected static final HashMap keyWords=new HashMap();
    protected static final HashMap symbols=new HashMap();
    public int ttype;

    public static final int TT_EOF=-1;
    public static final int TT_NUMBER=-2;
    public static final int TT_WORD=-3;
    public static final int TT_NEW=-4;
    public static final int TT_INT=-5;
    public static final int TT_CLASS=-6;
    public static final int TT_EXTENDS=-7;
    public static final int TT_RETURN=-8;
    public static final int TT_BEGIN_BLOCK=-9;  // {
    public static final int TT_END_BLOCK=-10;   // }
    public static final int TT_BEGIN_PAR=-11;   // (
    public static final int TT_END_PAR=-12;     // )
    public static final int TT_DOT=-13;         // .
    public static final int TT_COMMA=-14;       // ,
    public static final int TT_SEMICOLON=-15;   // ;
    public static final int TT_ASSIGN=-16;      // =
    public static final int TT_THIS=-17;
    
    static{
	keyWords.put("new",new Integer(TT_NEW));
	keyWords.put("int",new Integer(TT_INT));
	keyWords.put("class",new Integer(TT_CLASS));
	keyWords.put("extends",new Integer(TT_EXTENDS));
	keyWords.put("return",new Integer(TT_RETURN));
	keyWords.put("this",new Integer(TT_THIS));
	symbols.put("{",new Integer(TT_BEGIN_BLOCK));  
	symbols.put("}",new Integer(TT_END_BLOCK));  
	symbols.put("(",new Integer(TT_BEGIN_PAR));
	symbols.put(")",new Integer(TT_END_PAR));  
	symbols.put(".",new Integer(TT_DOT));
	symbols.put(",",new Integer(TT_COMMA));
	symbols.put(";",new Integer(TT_SEMICOLON));
	symbols.put("=",new Integer(TT_ASSIGN));
    }

    public StreamTokenizer(Reader r){
	super(r);
	resetSyntax();
    }

    public void resetSyntax(){
	super.resetSyntax();
	whitespaceChars('\u0000','\u0020');  
	slashSlashComments(true);
	slashStarComments(true);
    }

    public int next() throws IOException{
	String val;
	nextToken();
	if (super.ttype==super.TT_EOF)
	    ttype=TT_EOF;
	else if(super.ttype=='-' || super.ttype>='0' && super.ttype<='9'){ // number
	    val=String.valueOf((char)super.ttype);
	    super.resetSyntax();
	    do{
		nextToken();
		if(super.ttype>='0' && super.ttype<='9')
		    val=val.concat(String.valueOf((char)super.ttype));
		else
		    break;
	    }
	    while(true);
	    if(super.ttype>'\u0020') 
		pushBack();
	    if(val.length()==1 && val.charAt(0)=='-')
		ttype='-';
	    else{
		sval=val;
		ttype=TT_NUMBER;
	    }
	    resetSyntax();
	}
	else if (Character.isJavaIdentifierStart((char)super.ttype)){ // identifier
	    val=String.valueOf((char)super.ttype);
	    super.resetSyntax();
	    do{
		nextToken();
		if(Character.isJavaIdentifierPart((char)super.ttype))
		    val=val.concat(String.valueOf((char)super.ttype));
		else
		    break;
	    }
	    while(true);
	    if(super.ttype>'\u0020') 
		pushBack();
	    sval=val;
	    resetSyntax();
	    Object o=keyWords.get(sval);
	    if (o!=null) 
		ttype=((Integer)o).intValue();
	    else 
		ttype=TT_WORD;
	}
	else 
	    {
		Object o=symbols.get(String.valueOf((char)super.ttype));
		if (o!=null) 
		    ttype=((Integer)o).intValue();
		else
		    ttype=super.ttype;
	    }
	return ttype;
    }

    public static void main(String[] args)throws IOException{
	Reader r=new BufferedReader(new InputStreamReader(System.in));
	StreamTokenizer t=new StreamTokenizer(r);
	while (t.next()!=TT_EOF){
	    switch(t.ttype){
	    case TT_WORD:
		System.out.println("Word "+t.sval);
		break;
	    case TT_NUMBER:
		System.out.println("Number "+t.sval);
		break;
	    default:
		if (t.ttype<0) 
		    System.out.println("Symbol "+ t.ttype);
		else 
		    System.out.println("Char "+t.ttype);
	    }
	}
    }

}
