package FJsc.typechecker.typeassum;

import java.io.*;
import java.util.*;
import FJsc.typechecker.*;
import org.apache.bcel.generic.Type;
import org.apache.bcel.generic.ObjectType;


public class MethAssum extends TypeAssum{

    protected ObjectType target;
    protected Types args;
    protected Type type;
    protected MethSig methSig;

    public MethAssum(ObjectType target,Types args,Type type,MethSig methSig){
    	this.target=target;
    	this.args=args;
	this.type=type;
	this.methSig=methSig;
    }

    public MethAssum(Env ce,ObjectType target,String methName,Types args)throws TypecheckerException{
	this.target=target;
	this.args=args;
	this.methSig=new MethSig(methName); // the rest of the signature will be completed by method resolve
	this.type=resolve(ce); 
    }

    private Type resolve(Env ce)throws TypecheckerException{
	String className=target.toString();
	String methName=methSig.getName();
	Types ts;
	MethHeaders mhs;
	while(className!=null){
	    try{
		mhs=ce.getMeths(className);
		ts=mostSpecific(ce,applicable(ce,mhs));
		if(isMostSpecific(ts,ce,ce.getSuperclass(className))){
		    this.methSig=new MethSig(methName,ts);
		    return (Type) mhs.get(ts);
		}
		throw new NoMostSpecificException();
	    }
	    catch(NoApplicableException e){}
	    catch(NoMostSpecificException e){		
		throw new TypecheckerException("Ambiguous method invocation "+target+"."+methName+args);
	    }
	    className=ce.getSuperclass(className);
	}
	throw new TypecheckerException("No applicable methods found for "+target+"."+methName+args);
    }

    private List applicable(Env ce,MethHeaders mhs)throws NoApplicableException{
	List l=new ArrayList();
	MethSig ms;
	String methName=methSig.getName();
	Iterator it=mhs.keySet().iterator();
	while(it.hasNext()){ 
	    ms=(MethSig)it.next();
	    if(methName.equals(ms.getName()) && subtypeOf(ce,args,ms))
		l.add(ms);
	}
	if(l.isEmpty())
	    throw new NoApplicableException();
	return l;
    }

    private boolean isMostSpecific(Types formals,Env ce,String className)throws TypecheckerException{
	MethSig ms;
	String methName=methSig.getName();
	while(className!=null){
	    Iterator it=ce.getMeths(className).keySet().iterator();
	    while(it.hasNext()){
		ms=(MethSig)it.next();
		if(methName.equals(ms.getName()) && subtypeOf(ce,args,ms) && !subtypeOf(ce,formals,ms))
		    return false;
	    }
	    className=ce.getSuperclass(className);
	}
	return true;
    }

    public ObjectType getTarget(){
	return target;
    }

    public Types getArgs(){
	return args;
    }

    public Type getType(){
	return type;
    }

    public MethSig getSig(){
	return methSig;
    }

    public void entailedBy(Env ce)throws TypecheckerException{
	MethAssum ma=new MethAssum(ce,target,methSig.getName(),args);
	if(!type.equals(ma.type)|| !methSig.equals(ma.methSig))
	    throw new TypecheckerException("Cannot resolve method invocation "+target+"."+methSig.getName()+args);
    }

    public String toString(){
	return "meth("+target+","+methSig.getName()+","+args+") = "+type+" "+methSig;
    }

    public boolean equals(Object o){
	try{
	    return equals((MethAssum) o);
	}
	catch(ClassCastException e){
	    return false;
	}
    }

    public boolean equals(MethAssum ma){
	return target.equals(ma.target) && args.equals(ma.args) && methSig.equals(ma.methSig) && type.equals(ma.type);
    }

    public int hashCode(){
	return target.hashCode()+args.hashCode()+methSig.hashCode()+type.hashCode();
    }

}
